from os.path import abspath, dirname, exists, join
import sys

from .lexer import Lexer
from ply import lex, yacc

import inspect

DEBUG = False

# When using something like pyinstaller, the __file__ attribute isn't actually
# set correctly, so the parse file isn't able to be saved anywhere sensible.
# In these cases, just use a temporary directory, it doesn't take too long to
# generate the tables anyways...

if exists(dirname(__file__)):
    pickle_file = abspath(join(dirname(__file__), 'parsetab.dat'))
else:
    import tempfile

    fobj = tempfile.NamedTemporaryFile()
    pickle_file = fobj.name


if sys.version_info[0] < 3:

    def iteritems(d):
        return iter(d.iteritems())


else:

    def iteritems(d):
        return iter(d.items())


class HclParser(object):

    #
    # Tokens
    #

    tokens = (
        'BOOL',
        'FLOAT',
        'NUMBER',
        'COMMA',
        'COMMAEND',
        'IDENTIFIER',
        'EQUAL',
        'STRING',
        'MINUS',
        'LEFTBRACE',
        'RIGHTBRACE',
        'LEFTBRACKET',
        'RIGHTBRACKET',
        'PERIOD',
        'EPLUS',
        'EMINUS',
        'LEFTPAREN',
        'RIGHTPAREN',
        'QMARK',
        'COLON',
        'ASTERISK_PERIOD',
        'GT',
        'LT',
        'EQ',
        'NE',
        'LE',
        'GE',
    )

    #
    # Yacc parser section
    #
    def objectlist_flat(self, lt, replace):
        '''
            Similar to the dict constructor, but handles dups
            
            HCL is unclear on what one should do when duplicate keys are
            encountered. These comments aren't clear either:
            
            from decoder.go: if we're at the root or we're directly within
                             a list, decode into dicts, otherwise lists
                
            from object.go: there's a flattened list structure
        '''
        d = {}

        for k, v in lt:
            if k in d.keys() and not replace:
                if type(d[k]) is list:
                    d[k].append(v)
                else:
                    d[k] = [d[k], v]
            else:
                if isinstance(v, dict):
                    dd = d.setdefault(k, {})
                    for kk, vv in iteritems(v):
                        if type(dd) == list:
                            dd.append({kk: vv})
                        elif kk in dd.keys():
                            if hasattr(vv, 'items'):
                                for k2, v2 in iteritems(vv):
                                    dd[kk][k2] = v2
                            else:
                                d[k] = [dd, {kk: vv}]
                        else:
                            dd[kk] = vv
                else:
                    d[k] = v

        return d

    def p_top(self, p):
        "top : objectlist"
        if DEBUG:
            self.print_p(p)
        p[0] = self.objectlist_flat(p[1], True)

    def p_objectlist_0(self, p):
        "objectlist : objectitem"
        if DEBUG:
            self.print_p(p)
        p[0] = [p[1]]

    def p_objectlist_1(self, p):
        "objectlist : objectlist objectitem"
        if DEBUG:
            self.print_p(p)
        p[0] = p[1] + [p[2]]

    def p_objectlist_2(self, p):
        "objectlist : objectlist COMMA objectitem"
        if DEBUG:
            self.print_p(p)
        p[0] = p[1] + [p[3]]

    def p_object_0(self, p):
        "object : LEFTBRACE objectlist RIGHTBRACE"
        if DEBUG:
            self.print_p(p)
        p[0] = self.objectlist_flat(p[2], False)

    def p_object_1(self, p):
        "object : LEFTBRACE objectlist COMMA RIGHTBRACE"
        if DEBUG:
            self.print_p(p)
        p[0] = self.objectlist_flat(p[2], False)

    def p_object_2(self, p):
        "object : LEFTBRACE RIGHTBRACE"
        if DEBUG:
            self.print_p(p)
        p[0] = {}

    def p_objectkey_0(self, p):
        '''
        objectkey : IDENTIFIER
                  | STRING
        '''
        if DEBUG:
            self.print_p(p)
        p[0] = p[1]

    def p_objectbrackets_0(self, p):
        "objectbrackets : IDENTIFIER LEFTBRACKET objectkey RIGHTBRACKET"
        if DEBUG:
            self.print_p(p)
        p[0] = p[1] + p[2] + p[3] + p[4]

    def p_objectbrackets_1(self, p):
        '''
        objectbrackets : IDENTIFIER LEFTBRACKET objectkey RIGHTBRACKET PERIOD IDENTIFIER
                       | IDENTIFIER LEFTBRACKET NUMBER RIGHTBRACKET PERIOD IDENTIFIER
        '''
        if DEBUG:
            self.print_p(p)
        p[0] = p[1] + p[2] + str(p[3]) + p[4] + p[5] + p[6]

    def p_objectitem_0(self, p):
        '''
        objectitem : objectkey EQUAL number
                   | objectkey EQUAL BOOL
                   | objectkey EQUAL STRING
                   | objectkey EQUAL IDENTIFIER
                   | objectkey EQUAL object
                   | objectkey EQUAL list
                   | objectkey EQUAL objectbrackets
                   | objectkey EQUAL function
                   | objectkey COLON number
                   | objectkey COLON BOOL
                   | objectkey COLON STRING
                   | objectkey COLON IDENTIFIER
                   | objectkey COLON object
                   | objectkey COLON list
                   | objectkey COLON objectbrackets
        '''
        if DEBUG:
            self.print_p(p)
        p[0] = (p[1], p[3])

    def p_objectitem_1(self, p):
        "objectitem : block"
        if DEBUG:
            self.print_p(p)
        p[0] = p[1]

    def p_objectitem_2(self, p):
        '''
        objectitem : objectkey EQUAL objectkey QMARK objectkey COLON objectkey
                   | objectkey EQUAL objectkey QMARK objectkey COLON number
                   | objectkey EQUAL objectkey QMARK objectkey COLON BOOL
                   | objectkey EQUAL objectkey QMARK BOOL COLON objectkey
                   | objectkey EQUAL objectkey QMARK number COLON objectkey
                   | objectkey EQUAL objectkey QMARK number COLON number
                   | objectkey EQUAL objectkey QMARK number COLON BOOL
                   | objectkey EQUAL objectkey QMARK BOOL COLON number
                   | objectkey EQUAL objectkey QMARK BOOL COLON BOOL
                   | objectkey EQUAL booleanexp QMARK objectkey COLON objectkey
                   | objectkey EQUAL booleanexp QMARK objectkey COLON number
                   | objectkey EQUAL booleanexp QMARK objectkey COLON BOOL
                   | objectkey EQUAL booleanexp QMARK BOOL COLON objectkey
                   | objectkey EQUAL booleanexp QMARK number COLON objectkey
                   | objectkey EQUAL booleanexp QMARK number COLON number
                   | objectkey EQUAL booleanexp QMARK number COLON BOOL
                   | objectkey EQUAL booleanexp QMARK BOOL COLON number
                   | objectkey EQUAL booleanexp QMARK BOOL COLON BOOL
        '''
        if DEBUG:
            self.print_p(p)
        p[0] = (p[1], p[3] + p[4] + str(p[5]) + p[6] + str(p[7]))

    def p_operator_0(self, p):
        '''
        operator : EQ
                 | NE
                 | LT
                 | GT
                 | LE
                 | GE
        '''
        if DEBUG:
            self.print_p(p)
        p[0] = p[1]

    def p_booleanexp_0(self, p):
        '''
        booleanexp : objectkey operator objectkey
                   | objectkey operator number
                   | number operator objectkey
        '''
        if DEBUG:
            self.print_p(p)
        p[0] = str(p[1]) + p[2] + str(p[3])

    def p_block_0(self, p):
        "block : objectkey object"
        if DEBUG:
            self.print_p(p)
        p[0] = (p[1], p[2])

    def p_block_1(self, p):
        "block : objectkey block"
        if DEBUG:
            self.print_p(p)
        p[0] = (p[1], {p[2][0]: p[2][1]})

    def p_list_0(self, p):
        '''
        list : LEFTBRACKET listitems RIGHTBRACKET
             | LEFTBRACKET listitems COMMA RIGHTBRACKET
             | LEFTPAREN listitems RIGHTPAREN
             | LEFTPAREN listitems COMMA RIGHTPAREN
        '''
        if DEBUG:
            self.print_p(p)
        p[0] = p[2]

    def p_list_1(self, p):
        '''
        list : LEFTBRACKET RIGHTBRACKET
             | LEFTPAREN RIGHTPAREN
        '''
        if DEBUG:
            self.print_p(p)
        p[0] = []

    def p_list_2(self, p):
        '''
        list : LEFTPAREN LEFTBRACKET listitems RIGHTBRACKET PERIOD PERIOD PERIOD RIGHTPAREN
        '''
        if DEBUG:
            self.print_p(p)
        p[0] = p[1] + p[2] + p[3] + p[4] + p[5] + p[6] + p[7] + p[8]

    def p_function_0(self, p):
        '''
        function : IDENTIFIER list
        '''
        if DEBUG:
            self.print_p(p)
        p[2].insert(0, p[1])
        p[0] = p[2]

    def p_listitems_0(self, p):
        '''
        listitems : listitem
                  | function
                  | object COMMA
                  | objectkey COMMA
        '''
        if DEBUG:
            self.print_p(p)
        p[0] = [p[1]]

    def p_listitems_1(self, p):
        '''
        listitems : listitems COMMA listitem
                  | listitems COMMA function
                  | listitems COMMA objectkey
        '''
        if DEBUG:
            self.print_p(p)
        p[0] = p[1] + [p[3]]

    def p_listitems_2(self, p):
        '''
        listitems : object COMMA object
                  | object COMMA objectkey
                  | objectkey COMMA objectkey
                  | objectkey COMMA object
                  | objectkey COMMA list
                  | objectkey COMMA IDENTIFIER ASTERISK_PERIOD IDENTIFIER
        '''
        if DEBUG:
            self.print_p(p)
        p[0] = [p[1], p[3]]

    def p_listitems_3(self, p):
        '''
        listitems : objectkey list
        '''
        if DEBUG:
            self.print_p(p)
        p[2].insert(0, p[1])
        p[0] = p[2]

    def p_listitem_0(self, p):
        '''
        listitem : number
                 | object
                 | objectkey
                 | objectbrackets
        '''
        if DEBUG:
            self.print_p(p)
        p[0] = p[1]

    def p_listitem_1(self, p):
        '''
        listitem : IDENTIFIER ASTERISK_PERIOD IDENTIFIER
        '''
        if DEBUG:
            self.print_p(p)
        p[0] = p[1] + p[2] + p[3]

    def p_number_0(self, p):
        "number : int"
        if DEBUG:
            self.print_p(p)
        p[0] = p[1]

    def p_number_1(self, p):
        "number : float"
        if DEBUG:
            self.print_p(p)
        p[0] = float(p[1])

    def p_number_2(self, p):
        "number : int exp"
        if DEBUG:
            self.print_p(p)
        p[0] = float("{0}{1}".format(p[1], p[2]))

    def p_number_3(self, p):
        "number : float exp"
        if DEBUG:
            self.print_p(p)
        p[0] = float("{0}{1}".format(p[1], p[2]))

    def p_int_0(self, p):
        "int : MINUS int"
        if DEBUG:
            self.print_p(p)
        p[0] = -p[2]

    def p_int_1(self, p):
        "int : NUMBER"
        if DEBUG:
            self.print_p(p)
        p[0] = p[1]

    def p_float_0(self, p):
        "float : MINUS float"
        p[0] = p[2] * -1

    def p_float_1(self, p):
        "float : FLOAT"
        p[0] = p[1]

    def p_exp_0(self, p):
        "exp : EPLUS NUMBER"
        if DEBUG:
            self.print_p(p)
        p[0] = "e{0}".format(p[2])

    def p_exp_1(self, p):
        "exp : EMINUS NUMBER"
        if DEBUG:
            self.print_p(p)
        p[0] = "e-{0}".format(p[2])

    # useful for debugging the parser
    def print_p(self, p):
        if DEBUG:
            name = inspect.getouterframes(inspect.currentframe(), 2)[1][3]
            print(
                '%20s: %s' % (name, ' | '.join([str(p[i]) for i in range(0, len(p))]))
            )

    def p_error(self, p):
        # Derived from https://groups.google.com/forum/#!topic/ply-hack/spqwuM1Q6gM

        # Ugly hack since Ply doesn't provide any useful error information
        try:
            frame = inspect.currentframe()
            cvars = frame.f_back.f_locals
            expected = "; expected %s" % (
                ', '.join(cvars['actions'][cvars['state']].keys())
            )
        except:
            expected = ""

        if p is not None:
            msg = "Line %d, column %d: unexpected %s%s" % (
                p.lineno,
                p.lexpos,
                p.type,
                expected,
            )
        else:
            msg = "Unexpected end of file%s" % expected

        raise ValueError(msg)

    def __init__(self):
        self.yacc = yacc.yacc(
            module=self, debug=False, optimize=1, picklefile=pickle_file
        )

    def parse(self, s):
        return self.yacc.parse(s, lexer=Lexer())



"""Used to destroy :ref:`modules<runway-module>` with Runway.

.. danger:: Use extreme caution when using with ``CI`` or
            ``--tag <tag>...``. You will **not** be prompted before
            deletion. All modules (or those selected by tag) for the
            current environment will be **irrecoverably deleted**.

When run, the environment is determined from the current git branch
unless ``ignore_git_branch: true`` is specified in the
:ref:`Runway config file<runway-config>`. If the ``DEPLOY_ENVIRONMENT``
environment variable is set, it's value will be used. If neither the git
branch or environment variable are available, the directory name is used.
The environment identified here is used to determine the env/config files
to use. It is also used with options defined in the Runway config file
such as ``assume_role``, ``account_id``, etc. See
:ref:`Runway Config<runway-config>` for details on these options.

The user will be prompted to select which
:ref:`deployment(s)<runway-deployment>` and
:ref:`module(s)<runway-module>` to process unless there is only one
:ref:`deployment<runway-deployment>` and/or
:ref:`module<runway-module>`, the environment variable ``CI`` is set,
or the ``--tag <tag>...`` option provided is used. In which case, the
:ref:`deployment(s)<runway-deployment>` and :ref:`module(s)<runway-module>`
will be processed in sequence, in reverse of the order they are defined.

.. rubric:: Options

+--------------------+-------------------------------------------------+
| ``--tag <tag>...`` | | Select modules for processing by tag or tags. |
|                    |   This option can be specified                  |
|                    | | more than once to build a list of tags that   |
|                    |   are treated as "AND".                         |
|                    | | (ex. ``--tag <tag1> --tag <tag2>`` would      |
|                    |   select all modules with BOTH tags).           |
+--------------------+-------------------------------------------------+

.. rubric:: Example

.. code-block:: shell

    # manually select deployment(s) and module(s)
    $ runway destroy

    # select all modules with the tag 'app:example' AND 'my-tag'
    $ runway destroy --tag app:example --tag my-tag

    # process all deployment(s) and module(s)
    $ CI=1 runway destroy

"""
from ..modules_command import ModulesCommand


class Destroy(ModulesCommand):
    """Extend ModulesCommand with execute to run the destroy method."""

    def execute(self):
        """Destroy deployments."""
        self.run(deployments=None, command='destroy')



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import str
from past.builtins import basestring
from builtins import object
import copy
import hashlib
import logging
import string
from stacker.util import read_value_from_path
from stacker.variables import Variable

from troposphere import (
    Output,
    Parameter,
    Ref,
    Template,
)

from ..exceptions import (
    MissingVariable,
    UnresolvedVariable,
    UnresolvedVariables,
    ValidatorError,
    VariableTypeRequired,
    InvalidUserdataPlaceholder
)
from .variables.types import (
    CFNType,
    TroposphereType,
)

logger = logging.getLogger(__name__)

PARAMETER_PROPERTIES = {
    "default": "Default",
    "description": "Description",
    "no_echo": "NoEcho",
    "allowed_values": "AllowedValues",
    "allowed_pattern": "AllowedPattern",
    "max_length": "MaxLength",
    "min_length": "MinLength",
    "max_value": "MaxValue",
    "min_value": "MinValue",
    "constraint_description": "ConstraintDescription"
}


class CFNParameter(object):

    def __init__(self, name, value):
        """Wrapper around a value to indicate a CloudFormation Parameter.

        Args:
            name (str): the name of the CloudFormation Parameter
            value (str, list, int or bool): the value we're going to submit as
                a CloudFormation Parameter.

        """
        acceptable_types = [basestring, bool, list, int]
        acceptable = False
        for acceptable_type in acceptable_types:
            if isinstance(value, acceptable_type):
                acceptable = True
                if acceptable_type == bool:
                    logger.debug("Converting parameter %s boolean '%s' "
                                 "to string.", name, value)
                    value = str(value).lower()
                    break

                if acceptable_type == int:
                    logger.debug("Converting parameter %s integer '%s' "
                                 "to string.", name, value)
                    value = str(value)
                    break

        if not acceptable:
            raise ValueError(
                "CFNParameter (%s) value must be one of %s got: %s" % (
                    name, "str, int, bool, or list", value))

        self.name = name
        self.value = value

    def __repr__(self):
        return "CFNParameter({}: {})".format(self.name, self.value)

    def to_parameter_value(self):
        """Return the value to be submitted to CloudFormation"""
        return self.value

    @property
    def ref(self):
        return Ref(self.name)


def build_parameter(name, properties):
    """Builds a troposphere Parameter with the given properties.

    Args:
        name (string): The name of the parameter.
        properties (dict): Contains the properties that will be applied to the
            parameter. See:
            http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/parameters-section-structure.html

    Returns:
        :class:`troposphere.Parameter`: The created parameter object.
    """
    p = Parameter(name, Type=properties.get("type"))
    for name, attr in PARAMETER_PROPERTIES.items():
        if name in properties:
            setattr(p, attr, properties[name])
    return p


def validate_variable_type(var_name, var_type, value):
    """Ensures the value is the correct variable type.

    Args:
        var_name (str): The name of the defined variable on a blueprint.
        var_type (type): The type that the value should be.
        value (obj): The object representing the value provided for the
            variable

    Returns:
        object: Returns the appropriate value object. If the original value
            was of CFNType, the returned value will be wrapped in CFNParameter.

    Raises:
        ValueError: If the `value` isn't of `var_type` and can't be cast as
            that type, this is raised.
    """

    if isinstance(var_type, CFNType):
        value = CFNParameter(name=var_name, value=value)
    elif isinstance(var_type, TroposphereType):
        try:
            value = var_type.create(value)
        except Exception as exc:
            name = "{}.create".format(var_type.resource_name)
            raise ValidatorError(var_name, name, value, exc)
    else:
        if not isinstance(value, var_type):
            raise ValueError(
                "Value for variable %s must be of type %s. Actual "
                "type: %s." % (var_name, var_type, type(value))
            )

    return value


def validate_allowed_values(allowed_values, value):
    """Support a variable defining which values it allows.

    Args:
        allowed_values (Optional[list]): A list of allowed values from the
            variable definition
        value (obj): The object representing the value provided for the
            variable

    Returns:
        bool: Boolean for whether or not the value is valid.

    """
    # ignore CFNParameter, troposphere handles these for us
    if not allowed_values or isinstance(value, CFNParameter):
        return True

    return value in allowed_values


def resolve_variable(var_name, var_def, provided_variable, blueprint_name):
    """Resolve a provided variable value against the variable definition.

    Args:
        var_name (str): The name of the defined variable on a blueprint.
        var_def (dict): A dictionary representing the defined variables
            attributes.
        provided_variable (:class:`stacker.variables.Variable`): The variable
            value provided to the blueprint.
        blueprint_name (str): The name of the blueprint that the variable is
            being applied to.

    Returns:
        object: The resolved variable value, could be any python object.

    Raises:
        MissingVariable: Raised when a variable with no default is not
            provided a value.
        UnresolvedVariable: Raised when the provided variable is not already
            resolved.
        ValueError: Raised when the value is not the right type and cannot be
            cast as the correct type. Raised by
            :func:`stacker.blueprints.base.validate_variable_type`
        ValidatorError: Raised when a validator raises an exception. Wraps the
            original exception.
    """

    try:
        var_type = var_def["type"]
    except KeyError:
        raise VariableTypeRequired(blueprint_name, var_name)

    if provided_variable:
        if not provided_variable.resolved:
            raise UnresolvedVariable(blueprint_name, provided_variable)

        value = provided_variable.value
    else:
        # Variable value not provided, try using the default, if it exists
        # in the definition
        try:
            value = var_def["default"]
        except KeyError:
            raise MissingVariable(blueprint_name, var_name)

    # If no validator, return the value as is, otherwise apply validator
    validator = var_def.get("validator", lambda v: v)
    try:
        value = validator(value)
    except Exception as exc:
        raise ValidatorError(var_name, validator.__name__, value, exc)

    # Ensure that the resulting value is the correct type
    value = validate_variable_type(var_name, var_type, value)

    allowed_values = var_def.get("allowed_values")
    if not validate_allowed_values(allowed_values, value):
        message = (
            "Invalid value passed to '%s' in blueprint: %s. Got: '%s', "
            "expected one of %s"
        ) % (var_name, blueprint_name, value, allowed_values)
        raise ValueError(message)

    return value


def parse_user_data(variables, raw_user_data, blueprint_name):
    """Parse the given user data and renders it as a template

    It supports referencing template variables to create userdata
    that's supplemented with information from the stack, as commonly
    required when creating EC2 userdata files.

    For example:
        Given a raw_user_data string: 'open file ${file}'
        And a variables dictionary with: {'file': 'test.txt'}
        parse_user_data would output: open file test.txt

    Args:
        variables (dict): variables available to the template
        raw_user_data (str): the user_data to be parsed
        blueprint_name (str): the name of the blueprint

    Returns:
        str: The parsed user data, with all the variables values and
             refs replaced with their resolved values.

    Raises:
        InvalidUserdataPlaceholder: Raised when a placeholder name in
                                    raw_user_data is not valid.
                                    E.g ${100} would raise this.
        MissingVariable: Raised when a variable is in the raw_user_data that
                         is not given in the blueprint

    """
    variable_values = {}

    for key, value in variables.items():
        if type(value) is CFNParameter:
            variable_values[key] = value.to_parameter_value()
        else:
            variable_values[key] = value

    template = string.Template(raw_user_data)

    res = ""

    try:
        res = template.substitute(variable_values)
    except ValueError as exp:
        raise InvalidUserdataPlaceholder(blueprint_name, exp.args[0])
    except KeyError as key:
        raise MissingVariable(blueprint_name, key)

    return res


class Blueprint(object):

    """Base implementation for rendering a troposphere template.

    Args:
        name (str): A name for the blueprint.
        context (:class:`stacker.context.Context`): the context the blueprint
            is being executed under.
        mappings (dict, optional): Cloudformation Mappings to be used in the
            template.

    """

    def __init__(self, name, context, mappings=None, description=None):
        self.name = name
        self.context = context
        self.mappings = mappings
        self.outputs = {}
        self.reset_template()
        self.resolved_variables = None
        self.description = description

        if hasattr(self, "PARAMETERS") or hasattr(self, "LOCAL_PARAMETERS"):
            raise AttributeError("DEPRECATION WARNING: Blueprint %s uses "
                                 "deprecated PARAMETERS or "
                                 "LOCAL_PARAMETERS, rather than VARIABLES. "
                                 "Please update your blueprints. See https://"
                                 "stacker.readthedocs.io/en/latest/blueprints."
                                 "html#variables for aditional information."
                                 % name)

    def get_parameter_definitions(self):
        """Get the parameter definitions to submit to CloudFormation.

        Any variable definition whose `type` is an instance of `CFNType` will
        be returned as a CloudFormation Parameter.

        Returns:
            dict: parameter definitions. Keys are parameter names, the values
                are dicts containing key/values for various parameter
                properties.

        """
        output = {}
        for var_name, attrs in self.defined_variables().items():
            var_type = attrs.get("type")
            if isinstance(var_type, CFNType):
                cfn_attrs = copy.deepcopy(attrs)
                cfn_attrs["type"] = var_type.parameter_type
                output[var_name] = cfn_attrs
        return output

    def get_output_definitions(self):
        """Gets the output definitions.

        Returns:
            dict: output definitions. Keys are output names, the values
                are dicts containing key/values for various output
                properties.

        """
        return {k: output.to_dict() for k, output in
                self.template.outputs.items()}

    def get_required_parameter_definitions(self):
        """Returns all template parameters that do not have a default value.

        Returns:
            dict: dict of required CloudFormation Parameters for the blueprint.
                Will be a dictionary of <parameter name>: <parameter
                attributes>.

        """
        required = {}
        for name, attrs in self.get_parameter_definitions().items():
            if "Default" not in attrs:
                required[name] = attrs
        return required

    def get_parameter_values(self):
        """Return a dictionary of variables with `type` :class:`CFNType`.

        Returns:
            dict: variables that need to be submitted as CloudFormation
                Parameters. Will be a dictionary of <parameter name>:
                <parameter value>.

        """
        variables = self.get_variables()
        output = {}
        for key, value in variables.items():
            try:
                output[key] = value.to_parameter_value()
            except AttributeError:
                continue

        return output

    def setup_parameters(self):
        """Add any CloudFormation parameters to the template"""
        t = self.template
        parameters = self.get_parameter_definitions()

        if not parameters:
            logger.debug("No parameters defined.")
            return

        for name, attrs in parameters.items():
            p = build_parameter(name, attrs)
            t.add_parameter(p)

    def defined_variables(self):
        """Return a dictionary of variables defined by the blueprint.

        By default, this will just return the values from `VARIABLES`, but this
        makes it easy for subclasses to add variables.

        Returns:
            dict: variables defined by the blueprint

        """
        return copy.deepcopy(getattr(self, "VARIABLES", {}))

    def get_variables(self):
        """Return a dictionary of variables available to the template.

        These variables will have been defined within `VARIABLES` or
        `self.defined_variables`. Any variable value that contains a lookup
        will have been resolved.

        Returns:
            dict: variables available to the template

        Raises:

        """
        if self.resolved_variables is None:
            raise UnresolvedVariables(self.name)
        return self.resolved_variables

    def get_cfn_parameters(self):
        """Return a dictionary of variables with `type` :class:`CFNType`.

        Returns:
            dict: variables that need to be submitted as CloudFormation
                Parameters.

        """
        variables = self.get_variables()
        output = {}
        for key, value in variables.items():
            if hasattr(value, "to_parameter_value"):
                output[key] = value.to_parameter_value()
        return output

    def resolve_variables(self, provided_variables):
        """Resolve the values of the blueprint variables.

        This will resolve the values of the `VARIABLES` with values from the
        env file, the config, and any lookups resolved.

        Args:
            provided_variables (list of :class:`stacker.variables.Variable`):
                list of provided variables

        """
        self.resolved_variables = {}
        defined_variables = self.defined_variables()
        variable_dict = dict((var.name, var) for var in provided_variables)
        for var_name, var_def in defined_variables.items():
            value = resolve_variable(
                var_name,
                var_def,
                variable_dict.get(var_name),
                self.name
            )
            self.resolved_variables[var_name] = value

    def import_mappings(self):
        if not self.mappings:
            return

        for name, mapping in self.mappings.items():
            logger.debug("Adding mapping %s.", name)
            self.template.add_mapping(name, mapping)

    def reset_template(self):
        self.template = Template()
        self._rendered = None
        self._version = None

    def render_template(self):
        """Render the Blueprint to a CloudFormation template"""
        self.import_mappings()
        self.create_template()
        if self.description:
            self.set_template_description(self.description)
        self.setup_parameters()
        rendered = self.template.to_json(indent=self.context.template_indent)
        version = hashlib.md5(rendered.encode()).hexdigest()[:8]
        return (version, rendered)

    def to_json(self, variables=None):
        """Render the blueprint and return the template in json form.

        Args:
            variables (dict):
                Optional dictionary providing/overriding variable values.

        Returns:
            str: the rendered CFN JSON template
        """

        variables_to_resolve = []
        if variables:
            for key, value in variables.items():
                variables_to_resolve.append(Variable(key, value))
        for k in self.get_parameter_definitions():
            if not variables or k not in variables:
                # The provided value for a CFN parameter has no effect in this
                # context (generating the CFN template), so any string can be
                # provided for its value - just needs to be something
                variables_to_resolve.append(Variable(k, 'unused_value'))
        self.resolve_variables(variables_to_resolve)

        return self.render_template()[1]

    def read_user_data(self, user_data_path):
        """Reads and parses a user_data file.

        Args:
            user_data_path (str):
                path to the userdata file

        Returns:
            str: the parsed user data file

        """
        raw_user_data = read_value_from_path(user_data_path)

        variables = self.get_variables()

        return parse_user_data(variables, raw_user_data, self.name)

    def set_template_description(self, description):
        """Adds a description to the Template

        Args:
            description (str): A description to be added to the resulting
                template.

        """
        self.template.add_description(description)

    def add_output(self, name, value):
        """Simple helper for adding outputs.

        Args:
            name (str): The name of the output to create.
            value (str): The value to put in the output.
        """
        self.template.add_output(Output(name, Value=value))

    @property
    def requires_change_set(self):
        """Returns true if the underlying template has transforms."""
        return self.template.transform is not None

    @property
    def rendered(self):
        if not self._rendered:
            self._version, self._rendered = self.render_template()
        return self._rendered

    @property
    def version(self):
        if not self._version:
            self._version, self._rendered = self.render_template()
        return self._version

    def create_template(self):
        raise NotImplementedError



"""Serverless Hello World function."""
import json


def handler(event, context):  # pylint: disable=unused-argument
    """Return Serverless Hello World."""
    body = {
        "message": "Go Serverless v1.0! Your function executed successfully!",
        "input": event
    }

    response = {
        "statusCode": 200,
        "body": json.dumps(body)
    }

    return response

    # Use this code if you don't use the http event with the LAMBDA-PROXY
    # integration
    # return {
    #     "message": "Go Serverless v1.0! Your function executed!",
    #     "event": event
    # }



#!/usr/bin/env python
"""Module with k8s cluster resources."""
from __future__ import print_function

from troposphere import Export, Join, Output, Sub, ec2, eks, iam

import awacs.iam
from awacs.aws import Allow, Condition, PolicyDocument, Statement
from awacs.helpers.trust import make_simple_assume_policy
# Autogenerated condition classes trip up pylint
from awacs.aws import StringLike  # pylint: disable=no-name-in-module

from stacker.blueprints.base import Blueprint
from stacker.blueprints.variables.types import (
    CFNString, EC2SubnetIdList, EC2VPCId
)

IAM_POLICY_ARN_PREFIX = 'arn:aws:iam::aws:policy/'


class Cluster(Blueprint):
    """Stacker blueprint for creating k8s cluster resources."""

    VARIABLES = {
        'EksClusterName': {'type': CFNString,
                           'description': 'Name of the Kubernetes cluster',
                           'min_length': 2,
                           'max_length': 40},
        'EksSubnets': {'type': EC2SubnetIdList,
                       'description': 'Subnets where the Kubernetes cluster '
                                      'will live'},
        'EksVersion': {'type': CFNString,
                       'description': 'Kubernetes version'},
        'VPC': {'type': EC2VPCId,
                'description': 'VPC where the Kubernetes cluster will live'}
    }

    def create_template(self):
        """Create template (main function called by Stacker)."""
        template = self.template
        variables = self.get_variables()
        template.add_version('2010-09-09')
        template.add_description('Kubernetes Master via EKS - V1.0.0')

        # Resources
        ccpsecuritygroup = template.add_resource(
            ec2.SecurityGroup(
                'ClusterControlPlaneSecurityGroup',
                GroupDescription='Cluster communication with worker nodes',
                Tags=[
                    {'Key': Sub('kubernetes.io/cluster/${EksClusterName}'),
                     'Value': 'owned'},
                    {'Key': 'Product',
                     'Value': 'Kubernetes'},
                    {'Key': 'Project',
                     'Value': 'eks'},
                    {'Key': 'Name',
                     'Value': Sub('${EksClusterName}-sg-worker-nodes')}
                ],
                VpcId=variables['VPC'].ref
            )
        )
        template.add_output(
            Output(
                ccpsecuritygroup.title,
                Description='Cluster communication with worker nodes',
                Export=Export(
                    Sub('${AWS::StackName}-ControlPlaneSecurityGroup')
                ),
                Value=ccpsecuritygroup.ref()
            )
        )

        eksservicerole = template.add_resource(
            iam.Role(
                'EksServiceRole',
                AssumeRolePolicyDocument=make_simple_assume_policy(
                    'eks.amazonaws.com'
                ),
                ManagedPolicyArns=[
                    IAM_POLICY_ARN_PREFIX + i for i in [
                        'AmazonEKSClusterPolicy',
                        'AmazonEKSServicePolicy'
                    ]
                ],
                Policies=[
                    iam.Policy(
                        PolicyName='EksServiceRolePolicy',
                        PolicyDocument=PolicyDocument(
                            Statement=[
                                Statement(
                                    Action=[awacs.iam.CreateServiceLinkedRole,
                                            awacs.iam.PutRolePolicy],
                                    Condition=Condition(
                                        StringLike(
                                            'iam:AWSServiceName',
                                            'elasticloadbalancing.amazonaws.com'  # noqa
                                        )
                                    ),
                                    Effect=Allow,
                                    Resource=[
                                        Sub('arn:aws:iam::${AWS::AccountId}:role/'  # noqa
                                            'aws-service-role/'
                                            'elasticloadbalancing.amazonaws.com/'  # noqa
                                            'AWSServiceRoleForElasticLoadBalancing*')  # noqa
                                    ]
                                )
                            ]
                        )
                    )
                ]
            )
        )

        ekscluster = template.add_resource(
            eks.Cluster(
                'EksCluster',
                Name=variables['EksClusterName'].ref,
                Version=variables['EksVersion'].ref,
                RoleArn=eksservicerole.get_att('Arn'),
                ResourcesVpcConfig=eks.ResourcesVpcConfig(
                    SecurityGroupIds=[ccpsecuritygroup.ref()],
                    SubnetIds=variables['EksSubnets'].ref
                )
            )
        )
        template.add_output(
            Output(
                "%sName" % ekscluster.title,
                Description='EKS Cluster Name',
                Export=Export(
                    Sub("${AWS::StackName}-%sName" % ekscluster.title)
                ),
                Value=ekscluster.ref()
            )
        )
        template.add_output(
            Output(
                "%sEndpoint" % ekscluster.title,
                Description='EKS Cluster Endpoint',
                Export=Export(
                    Sub("${AWS::StackName}-%sEndpoint" % ekscluster.title)
                ),
                Value=ekscluster.get_att('Endpoint')
            )
        )

        # Additional Outputs
        template.add_output(
            Output(
                'VpcId',
                Description='EKS Cluster VPC Id',
                Export=Export(
                    Sub('${AWS::StackName}-VpcId')
                ),
                Value=variables['VPC'].ref
            )
        )
        template.add_output(
            Output(
                'Subnets',
                Description='EKS Cluster Subnets',
                Export=Export(
                    Sub('${AWS::StackName}-Subnets')
                ),
                Value=Join(',', variables['EksSubnets'].ref)
            )
        )


# Helper section to enable easy blueprint -> template generation
# (just run `python <thisfile>` to output the json)
if __name__ == "__main__":
    from stacker.context import Context
    print(Cluster('test', Context({"namespace": "test"}), None).to_json())



import json
from .parser import HclParser

import sys

if sys.version_info[0] < 3:

    def u(s):
        if isinstance(s, unicode):
            return s
        else:
            return unicode(s, 'utf-8')


else:

    def u(s):
        if isinstance(s, bytes):
            return s.decode('utf-8')
        else:
            return s


def isHcl(s):
    '''
        Detects whether a string is JSON or HCL
        
        :param s: String that may contain HCL or JSON
        
        :returns: True if HCL, False if JSON, raises ValueError
                  if neither
    '''
    for c in s:
        if c.isspace():
            continue

        if c == '{':
            return False
        else:
            return True

    raise ValueError("No HCL object could be decoded")


def load(fp):
    '''
        Deserializes a file-pointer like object into a python dictionary.
        The contents of the file must either be JSON or HCL.
        
        :param fp: An object that has a read() function
        
        :returns: Dictionary
    '''
    return loads(fp.read())


def loads(s):
    '''
        Deserializes a string and converts it to a dictionary. The contents
        of the string must either be JSON or HCL.
        
        :returns: Dictionary 
    '''
    s = u(s)
    if isHcl(s):
        return HclParser().parse(s)
    else:
        return json.loads(s)


def dumps(*args, **kwargs):
    '''Turns a dictionary into JSON, passthru to json.dumps'''
    return json.dumps(*args, **kwargs)



from __future__ import division
from __future__ import absolute_import
from __future__ import print_function


class Target(object):
    """A "target" is just a node in the stacker graph that does nothing, except
    specify dependencies. These can be useful as a means of logically grouping
    a set of stacks together that can be targeted with the `--targets` flag.
    """

    def __init__(self, definition):
        self.name = definition.name
        self.requires = definition.requires or []
        self.required_by = definition.required_by or []
        self.logging = False



"""
Runway Overview.

Usage:
  runway (test|preflight)
  runway (plan|taxi) [--tag <tag>...]
  runway (deploy|takeoff) [--tag <tag>...]
  runway (destroy|dismantle) [--tag <tag>...]
  runway init
  runway gen-sample [<samplename>]
  runway whichenv
  runway envvars
  runway run-aws <awscli-args>...
  runway run-python <filename>
  runway run-stacker <stacker-args>...
  runway tfenv (install|run) [<tfenv-args>...]
  runway kbenv (install|run) [<kbenv-args>...]
  runway -h | --help
  runway --version

Options:
  -h --help                         Show this screen.
  --version                         Show version.
  --tag <tag>...                    Select modules for processing by tag
                                    or tags. This option can be specified
                                    more than once to build a list of tags
                                    that are treated as "AND". (ex.
                                    "--tag <tag1> --tag <tag2>" would select
                                    all modules with BOTH tags).

Example:
  runway deploy --tag app:some-app --tag tier:web  # app:some-app AND tier:web
  runway destroy --tag app:some-app --tag tier:web  # app:some-app AND tier:web

Help:
  * Set the DEPLOY_ENVIRONMENT environment variable to set/override the
    autodetected environment. Autodetection is done from git branches in the
    form of ENV- (e.g. the dev environment is deployed from ENV-dev branch)
    falling back to name of the parent folder of the module.
  * All deploy commands (e.g. sls deploy, tf apply) will be run interactively
    unless the CI environment variable is set.

"""


import logging
import os

from docopt import docopt

from . import __version__ as version

from .commands.command_loader import find_command_class

LOGGER = logging.getLogger('runway')


def fix_hyphen_commands(raw_cli_arguments):
    """Update options to match their module names with underscores."""
    for i in ['gen-sample', 'run-aws', 'run-python', 'run-stacker']:
        raw_cli_arguments[i.replace('-', '_')] = raw_cli_arguments[i]
        raw_cli_arguments.pop(i)
    return raw_cli_arguments


def main():
    """Provide main CLI entrypoint."""
    if os.environ.get('DEBUG'):
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)
        # botocore info is spammy
        logging.getLogger('botocore').setLevel(logging.ERROR)

    cli_arguments = fix_hyphen_commands(docopt(__doc__, version=version))

    # at least one of these must be enabled, i.e. the value is 'True'... but unfortunately
    #  `docopts` doesn't give you the hierarchy... so given 'gen-sample cfn', there are
    #  TWO enabled items in the list, 'gen-sample' and 'cfn'
    possible_commands = [command for command, enabled in cli_arguments.items() if enabled]

    command_class = find_command_class(possible_commands)
    if command_class:
        command_class(cli_arguments).execute()
    else:
        LOGGER.error("class not found for command '%s'", possible_commands)



"""Execute the AWS CLI update-kubeconfig command."""
from __future__ import print_function
import os
import logging
import shutil

import six
import yaml

LOGGER = logging.getLogger(__name__)


def copy_template_to_env(path, env, region):
    """Copy k8s module template into new environment directory."""
    overlays_dir = os.path.join(path, 'overlays')
    template_dir = os.path.join(overlays_dir, 'template')
    env_dir = os.path.join(overlays_dir, env)
    if os.path.isdir(template_dir):
        if os.path.isdir(env_dir) or (
                os.path.isdir("%s-%s" % (env_dir, region))):
            LOGGER.info("Bootstrap of k8s module at \"%s\" skipped; module "
                        "already has a config for this environment", path)
        else:
            LOGGER.info("Copying overlay template at \"%s\" to new "
                        "environment directory \"%s\"", template_dir, env_dir)
            shutil.copytree(template_dir, env_dir, symlinks=True)
            # Update templated environment name in files
            for i in ['kustomization.yaml',
                      # namespace files can't be directly kustomized
                      'namespace.yaml']:
                templated_file_path = os.path.join(env_dir, i)
                if os.path.isfile(templated_file_path):
                    with open(templated_file_path, 'r') as stream:
                        filedata = stream.read()
                    if 'REPLACEMEENV' in filedata:
                        filedata = filedata.replace('REPLACEMEENV', env)
                        with open(templated_file_path, 'w') as stream:
                            stream.write(filedata)
    else:
        LOGGER.info("Skipping bootstrap of k8s module at \"%s\"; no template "
                    "directory present", path)


def create_runway_environments(provider, context, **kwargs):  # noqa pylint: disable=unused-argument
    """Copy k8s module templates into new environment directories.

    Args:
        provider (:class:`stacker.providers.base.BaseProvider`): provider
            instance
        context (:class:`stacker.context.Context`): context instance

    Returns: boolean for whether or not the hook succeeded.

    """
    LOGGER.info("Bootstrapping runway k8s modules, looking for unconfigured "
                "environments...")

    environment = kwargs['namespace']
    region = os.environ.get('AWS_DEFAULT_REGION')

    env_root = os.path.dirname(
        os.path.realpath(os.environ.get('RUNWAYCONFIG'))
    )
    with open(os.environ.get('RUNWAYCONFIG')) as data_file:
        runway_config = yaml.safe_load(data_file)

    for deployment in runway_config.get('deployments', []):
        for module in deployment.get('modules', []):
            if isinstance(module, six.string_types):
                path = module
            else:
                path = module.get('path')
            if path.endswith('.k8s'):
                copy_template_to_env(os.path.join(env_root, path),
                                     environment,
                                     region)
    return True



"""Manage versions and execute `Kubernetes`_ commands.

Runway's built-in kubectl version management ensure the correct version
is used for a given environment. Define a ``.kubectl-version`` file
in your k8s module and that version will be automatically downloaded &
used during Runway operations.

The ``tfenv`` subcommand supplements this functionality in 2 ways:

* The ``install`` option will download kubectl (e.g. for
  pre-seeding a deployment system)
* The ``run`` option will execute arbitrary kubectl commands

Examples:
  .. code-block:: shell

    $ runway kbenv install 1.14.5
    $ runway kbenv install  # retrieves version from .kubectl-version

    $ runway kbenv run -- get namespace

"""

import subprocess
import sys

from runway.env_mgr.kbenv import KBEnvManager
from ..runway_command import RunwayCommand
from ...util import strip_leading_option_delim


class KBEnv(RunwayCommand):
    """Extend RunwayCommand with execution of kbenv."""

    SKIP_FIND_CONFIG = True

    def execute(self):
        """Execute kbenv."""
        if self._cli_arguments.get('install'):
            if self._cli_arguments.get('<kbenv-args>'):
                KBEnvManager().install(self._cli_arguments.get('<kbenv-args>')[0])
            else:
                KBEnvManager().install()
        elif self._cli_arguments.get('run'):
            cmd_line_args = strip_leading_option_delim(
                self._cli_arguments.get('<kbenv-args>', [])
            )
            kb_bin = KBEnvManager().install()
            sys.exit(subprocess.call([kb_bin] + cmd_line_args))



#!/usr/bin/env python
"""Helper functions for bootstrap actions."""

import botocore


def bootstrap_value(value, context, **kwargs):
    """Return the bootstrap value on creation, otherwise the post_bootstrap.

    Format of value:
        <stack>::<bootstrap value>::<post_bootstrap_value>
    """
    try:
        stack_name, bootstrap_val, post_bootstrap_val = value.split("::")
    except ValueError:
        raise ValueError("Invalid value for bootstrap_value lookup: %s. Must "
                         "be in <stack>::<bootstrap value>::"
                         "<post_bootstrap val> format." % value)

    stack = next(i for i in context.get_stacks() if i.name == stack_name)
    try:
        stack_des = kwargs['provider'].cloudformation.describe_stacks(
            StackName=stack.fqn)['Stacks'][0]
    except botocore.exceptions.ClientError as exc:
        if "does not exist" not in str(exc):
            raise
        return bootstrap_val

    if kwargs['provider'].is_stack_completed(stack_des) or (
            kwargs['provider'].is_stack_in_progress(stack_des)):
        return post_bootstrap_val
    return bootstrap_val



"""Generate a sample :ref:`Runway module<runway-module>` directory.

The sample module is created in the current directory. If a directory
already exists with the name it tries to use, it will not create the
sample directory.

.. rubric:: Available Samples

+------------------+---------------------------------------------------+
|      Name        |  Description                                      |
+==================+===================================================+
| ``cdk-csharp``   | `AWS CDK`_ :ref:`module<runway-module>` using C#  |
+------------------+---------------------------------------------------+
| ``cdk-py``       | `AWS CDK`_ :ref:`module<runway-module>` using     |
|                  | Python                                            |
+------------------+---------------------------------------------------+
| ``cdk-tsc``      | `AWS CDK`_ :ref:`module<runway-module>` using     |
|                  | TypeScript                                        |
+------------------+---------------------------------------------------+
| ``cfn``          | `CloudFormation`_ :ref:`module<runway-module>`    |
|                  | stack with S3 bucket & DDB table (perfect for     |
|                  | storing Terraform backend state)                  |
+------------------+---------------------------------------------------+
| ``k8s-cfn-repo`` | `Kubernetes`_                                     |
|                  | :ref:`module<runway-module>` EKS cluster & sample |
|                  | app using CloudFormation                          |
+------------------+---------------------------------------------------+
| ``k8s-tf-repo``  | `Kubernetes`_                                     |
|                  | :ref:`module<runway-module>` EKS cluster & sample |
|                  | app using Terraform                               |
+------------------+---------------------------------------------------+
| ``sls-py``       | `Serverless Framework`_                           |
|                  | :ref:`module<runway-module>` using Python         |
+------------------+---------------------------------------------------+
| ``sls-tsc``      | `Serverless Framework`_                           |
|                  | :ref:`module<runway-module>` using TypeScript     |
+------------------+---------------------------------------------------+
| ``stacker``      | `Troposphere`_/`Stacker`_                         |
|                  | :ref:`module<runway-module>` identical the ``cfn``|
|                  | sample but with the template written in python    |
+------------------+---------------------------------------------------+
| ``tf``           | `Terraform`_ :ref:`module<runway-module>`         |
+------------------+---------------------------------------------------+

.. rubric:: Example

.. code-block:: shell

    # create a "sampleapp.cfn" sample module directory
    $ runway gen-sample cfn

    # create a "runway-sample-tfstate.cfn" sample module directory
    $ runway gen-sample stacker

    # create a "sampleapp.sls" sample module directory
    $ runway gen-sample sls-py

"""
from __future__ import print_function
import logging
import os
import shutil
import sys

from cfn_flip import to_yaml
from stacker.context import Context

from ..runway_command import RunwayCommand
from ...env_mgr.tfenv import get_latest_tf_version

LOGGER = logging.getLogger('runway')

ROOT = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))


def generate_tfstate_cfn_template():
    """Return rendered CFN template yaml."""
    from runway.blueprints.tf_state import TfState

    return to_yaml(TfState('test',
                           Context({"namespace": "test"}),
                           None).to_json())


def generate_sample_module(module_dir):
    """Generate skeleton sample module."""
    if os.path.isdir(module_dir):
        LOGGER.error("Error generating sample module -- directory %s "
                     "already exists!",
                     module_dir)
        sys.exit(1)
    os.mkdir(module_dir)


def generate_sample_k8s_cfn_repo(env_root):
    """Generate sample k8s infrastructure repo."""
    repo_dir = os.path.join(env_root, 'k8s-cfn-infrastructure')
    if os.path.isdir(repo_dir):
        LOGGER.error("Error generating sample repo -- directory %s "
                     "already exists!",
                     repo_dir)
        sys.exit(1)

    from runway.blueprints.k8s.k8s_master import Cluster
    from runway.blueprints.k8s.k8s_iam import Iam
    from runway.blueprints.k8s.k8s_workers import NodeGroup as WorkerNodeGroup

    shutil.copytree(
        os.path.join(ROOT,
                     'templates',
                     'k8s-cfn-repo'),
        repo_dir
    )
    os.rename(os.path.join(repo_dir, '_gitignore'),
              os.path.join(repo_dir, '.gitignore'))

    # Generate masters CFN templates from blueprints
    master_template_dir = os.path.join(repo_dir, 'k8s-master.cfn', 'templates')
    os.mkdir(master_template_dir)
    with open(os.path.join(master_template_dir, 'k8s_iam.yaml'), 'w') as stream:
        stream.write(to_yaml(Iam('test',
                                 Context({"namespace": "test"}),
                                 None).to_json()))
    with open(os.path.join(master_template_dir, 'k8s_master.yaml'), 'w') as stream:
        stream.write(to_yaml(Cluster('test',
                                     Context({"namespace": "test"}),
                                     None).to_json()))

    # Generate workers CFN template from blueprint
    worker_template_dir = os.path.join(repo_dir,
                                       'k8s-workers.cfn',
                                       'templates')
    os.mkdir(worker_template_dir)
    with open(os.path.join(worker_template_dir,
                           'k8s_workers.yaml'), 'w') as stream:
        stream.write(to_yaml(WorkerNodeGroup('test',
                                             Context({"namespace": "test"}),
                                             None).to_json()))

    LOGGER.info("Sample k8s infrastructure repo created at %s",
                repo_dir)
    LOGGER.info('(see its README for setup and deployment instructions)')


def generate_sample_k8s_tf_repo(env_root):
    """Generate sample k8s infrastructure repo."""
    repo_dir = os.path.join(env_root, 'k8s-tf-infrastructure')
    if os.path.isdir(repo_dir):
        LOGGER.error("Error generating sample repo -- directory %s "
                     "already exists!",
                     repo_dir)
        sys.exit(1)

    shutil.copytree(
        os.path.join(ROOT,
                     'templates',
                     'k8s-tf-repo'),
        repo_dir
    )
    os.rename(os.path.join(repo_dir, '_gitignore'),
              os.path.join(repo_dir, '.gitignore'))

    # Generate tfstate CFN template from blueprints
    tfstate_template_dir = os.path.join(repo_dir, 'tfstate.cfn', 'templates')
    os.mkdir(tfstate_template_dir)
    with open(os.path.join(tfstate_template_dir, 'tf_state.yml'), 'w') as stream:
        stream.write(generate_tfstate_cfn_template())

    LOGGER.info("Sample k8s infrastructure repo created at %s",
                repo_dir)
    LOGGER.info('(see its README for setup and deployment instructions)')


def generate_sample_sls_py_module(env_root, module_dir=None):
    """Generate skeleton Serverless python sample module."""
    if module_dir is None:
        module_dir = os.path.join(env_root, 'sampleapp.sls')

    if os.path.isdir(module_dir):
        LOGGER.error("Error generating sample module -- directory %s "
                     "already exists!",
                     module_dir)
        sys.exit(1)

    shutil.copytree(
        os.path.join(ROOT,
                     'templates',
                     'sls-py'),
        module_dir
    )
    os.rename(os.path.join(module_dir, '_gitignore'),
              os.path.join(module_dir, '.gitignore'))
    LOGGER.info("Sample Serverless module created at %s",
                module_dir)
    LOGGER.info('To finish its setup, change to the %s directory and execute '
                '"npm install" to generate its lockfile.', module_dir)


def generate_sample_sls_tsc_module(env_root, module_dir=None):
    """Generate skeleton Serverless TypeScript sample module."""
    if module_dir is None:
        module_dir = os.path.join(env_root, 'sampleapp.sls')
    generate_sample_module(module_dir)
    for i in ['package.json', 'serverless.yml', 'tsconfig.json',
              'webpack.config.js']:
        shutil.copyfile(
            os.path.join(ROOT,
                         'templates',
                         'sls-tsc',
                         i),
            os.path.join(module_dir, i),
        )
    os.mkdir(os.path.join(module_dir, 'src'))
    for i in ['handler.spec.ts', 'handler.ts']:
        shutil.copyfile(
            os.path.join(ROOT,
                         'templates',
                         'sls-tsc',
                         'src',
                         i),
            os.path.join(module_dir, 'src', i),
        )
    LOGGER.info("Sample Serverless TypeScript module created at %s",
                module_dir)


def generate_sample_cdk_tsc_module(env_root, module_dir=None):
    """Generate skeleton CDK TS sample module."""
    if module_dir is None:
        module_dir = os.path.join(env_root, 'sampleapp.cdk')
    generate_sample_module(module_dir)
    for i in ['.npmignore', 'cdk.json', 'package.json', 'runway.module.yml',
              'tsconfig.json', 'README.md']:
        shutil.copyfile(
            os.path.join(ROOT,
                         'templates',
                         'cdk-tsc',
                         i),
            os.path.join(module_dir, i),
        )
    for i in [['bin', 'sample.ts'], ['lib', 'sample-stack.ts']]:
        os.mkdir(os.path.join(module_dir, i[0]))
        shutil.copyfile(
            os.path.join(ROOT,
                         'templates',
                         'cdk-tsc',
                         i[0],
                         i[1]),
            os.path.join(module_dir, i[0], i[1]),
        )
    with open(os.path.join(module_dir, '.gitignore'), 'w') as stream:
        stream.write('*.js\n')
        stream.write('*.d.ts\n')
        stream.write('node_modules\n')
    LOGGER.info("Sample CDK module created at %s", module_dir)
    LOGGER.info('To finish its setup, change to the %s directory and execute '
                '"npm install" to generate its lockfile.', module_dir)


def generate_sample_cdk_cs_module(env_root, module_dir=None):
    """Generate skeleton CDK C# sample module."""
    if module_dir is None:
        module_dir = os.path.join(env_root, 'sampleapp.cdk')
    generate_sample_module(module_dir)
    for i in ['add-project.hook.d.ts', 'cdk.json', 'package.json',
              'runway.module.yml', 'README.md']:
        shutil.copyfile(
            os.path.join(ROOT,
                         'templates',
                         'cdk-csharp',
                         i),
            os.path.join(module_dir, i),
        )
    shutil.copyfile(
        os.path.join(ROOT,
                     'templates',
                     'cdk-csharp',
                     'dot_gitignore'),
        os.path.join(module_dir, '.gitignore'),
    )
    os.mkdir(os.path.join(module_dir, 'src'))
    shutil.copyfile(
        os.path.join(ROOT,
                     'templates',
                     'cdk-csharp',
                     'src',
                     'HelloCdk.sln'),
        os.path.join(module_dir, 'src', 'HelloCdk.sln'),
    )
    os.mkdir(os.path.join(module_dir, 'src', 'HelloCdk'))
    for i in ['HelloCdk.csproj', 'HelloConstruct.cs', 'HelloStack.cs',
              'Program.cs']:
        shutil.copyfile(
            os.path.join(ROOT,
                         'templates',
                         'cdk-csharp',
                         'src',
                         'HelloCdk',
                         i),
            os.path.join(module_dir, 'src', 'HelloCdk', i),
        )
    LOGGER.info("Sample C# CDK module created at %s", module_dir)
    LOGGER.info('To finish its setup, change to the %s directory and execute '
                '"npm install" to generate its lockfile.', module_dir)


def generate_sample_cdk_py_module(env_root, module_dir=None):
    """Generate skeleton CDK python sample module."""
    if module_dir is None:
        module_dir = os.path.join(env_root, 'sampleapp.cdk')
    generate_sample_module(module_dir)
    os.mkdir(os.path.join(module_dir, 'hello'))
    for i in ['hello/__init__.py', 'hello/hello_construct.py',
              'hello/hello_stack.py', '.gitignore', 'app.py', 'cdk.json',
              'package.json', 'Pipfile', 'Pipfile.lock', 'runway.module.yml']:
        shutil.copyfile(
            os.path.join(ROOT,
                         'templates',
                         'cdk-py',
                         i),
            os.path.join(module_dir, i),
        )
    with open(os.path.join(module_dir, '.gitignore'), 'w') as stream:
        stream.write('node_modules')
    LOGGER.info("Sample CDK module created at %s", module_dir)
    LOGGER.info('To finish its setup, change to the %s directory and execute '
                '"npm install" to generate its lockfile.', module_dir)


def generate_sample_cfn_module(env_root, module_dir=None):
    """Generate skeleton CloudFormation sample module."""
    if module_dir is None:
        module_dir = os.path.join(env_root, 'sampleapp.cfn')
    generate_sample_module(module_dir)
    for i in ['stacks.yaml', 'dev-us-east-1.env']:
        shutil.copyfile(
            os.path.join(ROOT,
                         'templates',
                         'cfn',
                         i),
            os.path.join(module_dir, i)
        )
    os.mkdir(os.path.join(module_dir, 'templates'))
    with open(os.path.join(module_dir,
                           'templates',
                           'tf_state.yml'), 'w') as stream:
        stream.write(generate_tfstate_cfn_template())
    LOGGER.info("Sample CloudFormation module created at %s",
                module_dir)


def generate_sample_stacker_module(env_root, module_dir=None):
    """Generate skeleton Stacker sample module."""
    if module_dir is None:
        module_dir = os.path.join(env_root,
                                  'runway-sample-tfstate.cfn')
    generate_sample_module(module_dir)
    for i in ['stacks.yaml', 'dev-us-east-1.env']:
        shutil.copyfile(
            os.path.join(ROOT,
                         'templates',
                         'stacker',
                         i),
            os.path.join(module_dir, i)
        )
    os.mkdir(os.path.join(module_dir, 'tfstate_blueprints'))
    for i in ['__init__.py', 'tf_state.py']:
        shutil.copyfile(
            os.path.join(ROOT,
                         'blueprints',
                         i),
            os.path.join(module_dir, 'tfstate_blueprints', i)
        )
    os.chmod(  # make blueprint executable
        os.path.join(module_dir, 'tfstate_blueprints', 'tf_state.py'),
        os.stat(os.path.join(module_dir,
                             'tfstate_blueprints',
                             'tf_state.py')).st_mode | 0o0111
    )
    LOGGER.info("Sample Stacker module created at %s",
                module_dir)


def generate_sample_tf_module(env_root, module_dir=None):
    """Generate skeleton Terraform sample module."""
    if module_dir is None:
        module_dir = os.path.join(env_root, 'sampleapp.tf')
    generate_sample_module(module_dir)
    for i in ['backend-us-east-1.tfvars', 'dev-us-east-1.tfvars', 'main.tf']:
        shutil.copyfile(
            os.path.join(ROOT,
                         'templates',
                         'terraform',
                         i),
            os.path.join(module_dir, i),
        )
    tf_ver_template = os.path.join(ROOT,
                                   'templates',
                                   'terraform',
                                   '.terraform-version')
    if os.path.isfile(tf_ver_template):
        shutil.copyfile(
            tf_ver_template,
            os.path.join(module_dir, '.terraform-version'),
        )
    else:  # running directly from git
        latest_tf_ver = get_latest_tf_version()
        with open(os.path.join(module_dir,
                               '.terraform-version'), 'w') as stream:
            stream.write(latest_tf_ver)

    LOGGER.info("Sample Terraform app created at %s",
                module_dir)


class GenSample(RunwayCommand):
    """Extend Base with execute to run the module generators."""

    SKIP_FIND_CONFIG = True

    def execute(self):
        """Run selected module generator."""
        if self._cli_arguments.get('<samplename>') == 'cfn':
            generate_sample_cfn_module(self.env_root)
        elif self._cli_arguments.get('<samplename>') == 'sls-py':
            generate_sample_sls_py_module(self.env_root)
        elif self._cli_arguments.get('<samplename>') == 'sls-tsc':
            generate_sample_sls_tsc_module(self.env_root)
        elif self._cli_arguments.get('<samplename>') == 'stacker':
            generate_sample_stacker_module(self.env_root)
        elif self._cli_arguments.get('<samplename>') == 'tf':
            generate_sample_tf_module(self.env_root)
        elif self._cli_arguments.get('<samplename>') == 'k8s-cfn-repo':
            generate_sample_k8s_cfn_repo(self.env_root)
        elif self._cli_arguments.get('<samplename>') == 'k8s-tf-repo':
            generate_sample_k8s_tf_repo(self.env_root)
        elif self._cli_arguments.get('<samplename>') == 'cdk-tsc':
            generate_sample_cdk_tsc_module(self.env_root)
        elif self._cli_arguments.get('<samplename>') == 'cdk-py':
            generate_sample_cdk_py_module(self.env_root)
        elif self._cli_arguments.get('<samplename>') == 'cdk-csharp':
            generate_sample_cdk_cs_module(self.env_root)
        else:
            LOGGER.info("Available samples to generate:")
            for i in ['cfn', 'sls-tsc', 'sls-py', 'tf', 'k8s-cfn-repo',
                      'k8s-tf-repo', 'stacker', 'cdk-tsc', 'cdk-py',
                      'cdk-csharp']:
                print(i)



"""Package setup file."""
from setuptools import setup

setup(name='src',
      version='1.0.0',
      description='Lambda function package',
      packages=['src'],
      install_requires=[
          'PyYAML',
      ],
      zip_safe=False)



"""Creates a sample :ref:`runway-config` in the current directory.

If a :ref:`Runway config file<runway-config>` is already present, no
action is taken.

Example:
  .. code-block:: shell

    $ runway init

.. rubric:: Sample Runway Config File
.. code-block:: yaml

    ---
    # See full syntax at https://docs.onica.com/projects/runway/en/latest/
    deployments:
      - modules:
          - nameofmyfirstmodulefolder
          - nameofmysecondmodulefolder
          # - etc...
        regions:
          - us-east-1

"""
from __future__ import print_function

import os
import sys

from ..runway_command import RunwayCommand


class Init(RunwayCommand):
    """Extend Base with init command."""

    SKIP_FIND_CONFIG = True

    def execute(self):  # pylint: disable=no-self-use
        """Generate runway.yml."""
        if os.path.isfile('runway.yml'):
            print('Runway config already present')
            sys.exit(1)
        with open('runway.yml', 'w') as stream:
            stream.write("""---
# See full syntax at https://docs.onica.com/projects/runway/en/latest/
deployments:
  - modules:
      - nameofmyfirstmodulefolder
      - nameofmysecondmodulefolder
      # - etc...
    regions:
      - us-east-1
""")
        print('runway.yml generated')
        print('See additional getting started information at '
              'https://docs.onica.com/projects/runway/en/latest/how_to_use.html')  # noqa pylint: disable=line-too-long



"""Base class for commands that need to parse the Runway config."""
from typing import List, Optional, Union  # pylint: disable=unused-import
import os
import logging

from ..config import Config


class BaseCommand(object):
    """Base class for commands."""

    SKIP_FIND_CONFIG = False  # set to true for commands that don't need config

    def __init__(self,
                 cli_arguments,  # type: List[Union[str, list, bool]]
                 env_root=None,  # type: Optional[str]
                 runway_config_dir=None  # type: Optional[str]
                 # pylint only complains for python2
                 ):  # pylint: disable=bad-continuation
        # type: (...) -> None
        """Initialize base class."""
        self._cli_arguments = cli_arguments

        if env_root is None:
            self.env_root = os.getcwd()
        else:
            self.env_root = env_root

        if self.SKIP_FIND_CONFIG:
            self.runway_config_path = None
            self._runway_config = None
            return

        # This may be invoked from a module directory in an environment;
        # account for that here if necessary
        try:
            # Disable other runway logging so the only response is the env name
            logging.getLogger('runway').setLevel(logging.CRITICAL)
            self.runway_config_path = Config.find_config_file(runway_config_dir or
                                                              self.env_root)
        except SystemExit:
            logging.getLogger('runway').setLevel(logging.WARN)
            self.env_root = os.path.dirname(os.getcwd())
            self.runway_config_path = Config.find_config_file(self.env_root)
        logging.getLogger('runway').setLevel(logging.INFO)
        self._runway_config = None

    @property
    def runway_config(self):
        # type: () -> Config
        """Return parsed runway.yml."""
        if not self._runway_config:
            self._runway_config = Config.load_from_file(
                self.runway_config_path
            )
        return self._runway_config

    def execute(self):
        # type: () -> None
        """Execute the command."""
        raise NotImplementedError('execute must be implimented for '
                                  'subclasses of BaseCommand.')



"""Used to plan actions by comparing what is live and what is defined locally.

.. note:: Currently only supported for `AWS CDK`_, `CloudFormation`_,
          `Terraform`_, and `Troposphere`_.

When run, the environment is determined from the current git branch
unless ``ignore_git_branch: true`` is specified in the
:ref:`Runway config file<runway-config>`. If the ``DEPLOY_ENVIRONMENT``
environment variable is set, it's value will be used. If neither the git
branch or environment variable are available, the directory name is used.
The environment identified here is used to determine the env/config files
to use. It is also used with options defined in the Runway config file
such as ``assume_role``, ``account_id``, etc. See
:ref:`Runway Config<runway-config>` for details on these options.

The user will be prompted to select which
:ref:`deployment(s)<runway-deployment>` and
:ref:`module(s)<runway-module>` to process unless there is only one
:ref:`deployment<runway-deployment>` and/or
:ref:`module<runway-module>`, the environment variable ``CI`` is set,
or the ``--tag <tag>...`` option provided is used. In which case, the
:ref:`deployment(s)<runway-deployment>` and :ref:`module(s)<runway-module>`
will be processed in sequence, in the order they are defined.

.. rubric:: Options

+--------------------+-------------------------------------------------+
| ``--tag <tag>...`` | | Select modules for processing by tag or tags. |
|                    |   This option can be specified                  |
|                    | | more than once to build a list of tags that   |
|                    |   are treated as "AND".                         |
|                    | | (ex. ``--tag <tag1> --tag <tag2>`` would      |
|                    |   select all modules with BOTH tags).           |
+--------------------+-------------------------------------------------+

.. rubric:: Equivalent To

There are the native commands that are used:

- ``cdk diff`` - https://docs.aws.amazon.com/cdk/latest/guide/tools.html
- ``stacker diff`` -
  https://stacker.readthedocs.io/en/stable/commands.html#diff
- ``terraform plan`` - https://www.terraform.io/docs/commands/plan.html

.. rubric:: Example

.. code-block:: shell

    $ runway plan

"""
from ..modules_command import ModulesCommand


class Plan(ModulesCommand):
    """Extend ModulesCommand with execute to run the plan method."""

    def execute(self):
        """Generate plans."""
        self.run(deployments=None, command='plan')



"""Packaging settings."""
import sys
from codecs import open as codecs_open
from os.path import abspath, dirname, join

from setuptools import find_packages, setup

from src.runway import __version__


THIS_DIR = abspath(dirname(__file__))
with codecs_open(join(THIS_DIR, 'README.rst'), encoding='utf-8') as readfile:
    LONG_DESCRIPTION = readfile.read()


INSTALL_REQUIRES = [
    'Send2Trash',
    'awacs',  # for embedded hooks
    # awscli included for embedded hooks and aws subcommand
    # version set to match stacker requirement and include awscli fix #4182
    'awscli>=1.16.191<2.0',
    'botocore>=1.12.111',  # matching awscli/boto3 requirement
    'boto3>=1.9.111<2.0',  # matching stacker requirement
    'cfn-lint',
    'docopt',
    'requests',
    'future',
    # embedded pyhcl is 0.3.12
    # with the LICENSE file added to its root folder
    # and the following patches applied
    # https://github.com/virtuald/pyhcl/pull/57
    'pyhcl~=0.3',
    'pyOpenSSL',  # For embedded hook & associated script usage
    'six',
    'typing;python_version<"3.5"',
    'yamllint',
    'zgitignore',  # for embedded hooks
    # embedded stacker is v1.7.0
    # with the LICENSE file added to its root folder
    # and the following patches applied
    # https://github.com/cloudtools/stacker/pull/731 (CAPABILITY_AUTO_EXPAND)
    # https://github.com/cloudtools/stacker/pull/744 (diffs via CFN changesets)
    # and the following files/folders deleted:
    #   * tests
    #   * blueprints/testutil.py
    # and the stacker & stacker.cmd scripts adapted with EMBEDDED_LIB_PATH
    'stacker~=1.7',
    # stacker's troposphere dep is more loose, but we need to ensure we use a
    # sufficiently recent version for compatibility embedded blueprints
    'troposphere>=2.4.2',
    # botocore pins its urllib3 dependency like this, so we need to do the
    # same to ensure v1.25+ isn't pulled in by pip
    'urllib3>=1.20,<1.25',
]

# ensuring pyyaml dep matches awscli
if sys.version_info[:2] == (2, 6):
    INSTALL_REQUIRES.append('PyYAML>=3.10,<=3.13')
    INSTALL_REQUIRES.append('cfn_flip<=1.2.0')  # 1.2.1+ require PyYAML 4.1+
else:
    INSTALL_REQUIRES.append('PyYAML>=4.1,<=5.1')
    INSTALL_REQUIRES.append('cfn_flip>=1.2.1')


setup(
    name='runway',
    version=__version__,
    description='Simplify infrastructure/app testing/deployment',
    long_description=LONG_DESCRIPTION,
    url='https://github.com/onicagroup/runway',
    author='Onica Group LLC',
    author_email='opensource@onica.com',
    license='Apache License 2.0',
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Utilities',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    python_requires='>=2.6',
    keywords='cli',
    # exclude=['docs', 'tests*'],
    packages=find_packages(where='src'),
    package_dir={"": "src"},
    install_requires=INSTALL_REQUIRES,
    extras_require={
        'test': ['flake8', 'pep8-naming', 'flake8-docstrings',
                 'mock~=3.0.5', 'pylint',
                 # python3 flake8-docstrings fails with pydocstyle 4:
                 # https://github.com/PyCQA/pydocstyle/issues/375
                 # newer versions do not support python2:
                 # https://github.com/PyCQA/pydocstyle/pull/374
                 'pydocstyle<4.0.0'],
    },
    entry_points={
        'console_scripts': [
            'runway=runway.cli:main',
        ],
    },
    scripts=['scripts/stacker-runway', 'scripts/stacker-runway.cmd',
             'scripts/tf-runway', 'scripts/tf-runway.cmd',
             'scripts/tfenv-runway', 'scripts/tfenv-runway.cmd'],
    include_package_data=True,  # needed for templates,blueprints,hooks
    test_suite='tests'
)



"""Execute the embedded copy of `Stacker`_.

This command allows direct access to Runway's CloudFormation management
tool, Stacker.

Example:
  .. code-block:: shell

    $ runway run-stacker -- build example.env example.yaml

"""

import logging
import sys

from ..runway_command import RunwayCommand
from ...util import get_embedded_lib_path, strip_leading_option_delim

LOGGER = logging.getLogger('runway')


class RunStacker(RunwayCommand):
    """Extend RunwayCommand with execution of Stacker."""

    SKIP_FIND_CONFIG = True

    def execute(self):
        """Execute stacker."""
        cmd_line_args = strip_leading_option_delim(
            self._cli_arguments.get('<stacker-args>', [])
        )
        lib_path = get_embedded_lib_path()
        sys.argv = ['stacker'] + cmd_line_args
        sys.path.insert(1, lib_path)
        from stacker.logger import setup_logging
        from stacker.commands import Stacker
        stacker = Stacker(setup_logging=setup_logging)
        args = stacker.parse_args(cmd_line_args)
        stacker.configure(args)
        args.run(args)



"""Utility functions for S3."""
import logging
import tempfile
import os
import zipfile
import boto3

from boto3.s3.transfer import S3Transfer
from botocore.exceptions import ClientError

LOGGER = logging.getLogger('runway')


def _get_client(session=None, region=None):
    """Get S3 boto client."""
    return session.client('s3') if session else boto3.client('s3', region_name=region)


def _get_resource(session=None, region=None):
    """Get S3 boto resource."""
    return session.resource('s3') if session else boto3.resource('s3', region_name=region)


def purge_and_delete_bucket(bucket_name, region='us-east-1', session=None):
    """Delete all objects and versions in bucket, then delete bucket."""
    purge_bucket(bucket_name, region, session)
    delete_bucket(bucket_name, region, session)


def purge_bucket(bucket_name, region='us-east-1', session=None):
    """Delete all objects and versions in bucket."""
    if does_bucket_exist(bucket_name, region, session):
        s3_resource = _get_resource(session, region)
        bucket = s3_resource.Bucket(bucket_name)
        bucket.object_versions.delete()
    else:
        LOGGER.warning('Bucket "%s" does not exist in region "%s"', bucket_name, region)


def delete_bucket(bucket_name, region='us-east-1', session=None):
    """Delete bucket."""
    if does_bucket_exist(bucket_name, region, session):
        LOGGER.info('Deleting bucket "%s"', bucket_name)
        s3_resource = _get_resource(session, region)
        bucket = s3_resource.Bucket(bucket_name)
        bucket.delete()
    else:
        LOGGER.warning('Bucket "%s" does not exist in region "%s"', bucket_name, region)


def does_bucket_exist(bucket_name, region='us-east-1', session=None):
    """Check if bucket exists in S3."""
    s3_resource = _get_resource(session, region)

    try:
        s3_resource.meta.client.head_bucket(Bucket=bucket_name)
        return True
    except ClientError as exc:
        if exc.response['Error']['Message'] == 'Not Found':
            LOGGER.info("Bucket \"%s\" does not exist", bucket_name)
            return False
        if exc.response['Error']['Message'] == 'Forbidden':
            LOGGER.exception("Access denied for bucket %s (permissions?)",
                             bucket_name)
            raise


def ensure_bucket_exists(bucket_name, region='us-east-1', session=None):
    """Ensure S3 bucket exists."""
    if not does_bucket_exist(bucket_name, region, session):
        LOGGER.info("Bucket \"%s\" does not exist, creating...", bucket_name)
        s3_client = _get_client(session, region)
        if region == 'us-east-1':
            create_bucket_opts = {}
        else:
            create_bucket_opts = {
                'CreateBucketConfiguration': {
                    'LocationConstraint': region
                }
            }
        s3_client.create_bucket(Bucket=bucket_name, **create_bucket_opts)

        # enable default encryption
        s3_client.put_bucket_encryption(Bucket=bucket_name,
                                        ServerSideEncryptionConfiguration={
                                            'Rules': [{
                                                'ApplyServerSideEncryptionByDefault': {
                                                    'SSEAlgorithm': 'AES256'
                                                }
                                            }]
                                        })


def does_s3_object_exist(bucket, key, session=None, region='us-east-1'):
    """Determine if object exists on s3."""
    s3_resource = _get_resource(session, region)

    try:
        s3_resource.Object(bucket, key).load()
    except ClientError as exc:
        if exc.response['Error']['Code'] == '404':
            return False
        raise
    return True


def upload(bucket, key, filename, session=None):
    """Upload file to S3 bucket."""
    s3_client = _get_client(session)
    LOGGER.info('Uploading %s to %s/%s', filename, bucket, key)
    s3_client.upload_file(filename, bucket, key)


def download(bucket, key, file_path, session=None):
    """Download a file from S3 to the given path."""
    s3_client = _get_client(session)

    transfer = S3Transfer(s3_client)
    transfer.download_file(bucket, key, file_path)
    return file_path


def download_and_extract_to_mkdtemp(bucket, key, session=None):
    """Download zip archive and extract it to temporary directory."""
    filedes, temp_file = tempfile.mkstemp()
    os.close(filedes)
    download(bucket, key, temp_file, session)

    output_dir = tempfile.mkdtemp()
    zip_ref = zipfile.ZipFile(temp_file, 'r')
    zip_ref.extractall(output_dir)
    zip_ref.close()
    os.remove(temp_file)
    return output_dir


def get_matching_s3_objects(bucket, prefix="", suffix="", session=None):
    """
    Generate objects in an S3 bucket.

    :param bucket: Name of the S3 bucket.
    :param prefix: Only fetch objects whose key starts with
        this prefix (optional).
    :param suffix: Only fetch objects whose keys end with
        this suffix (optional).
    """
    s3_client = _get_client(session)
    paginator = s3_client.get_paginator("list_objects_v2")

    kwargs = {'Bucket': bucket}

    # We can pass the prefix directly to the S3 API.  If the user has passed
    # a tuple or list of prefixes, we go through them one by one.
    if isinstance(prefix, str):
        prefixes = (prefix, )
    else:
        prefixes = prefix

    for key_prefix in prefixes:
        kwargs["Prefix"] = key_prefix

        for page in paginator.paginate(**kwargs):
            try:
                contents = page["Contents"]
            except KeyError:
                return

            for obj in contents:
                key = obj["Key"]
                if key.endswith(suffix):
                    yield obj


def get_matching_s3_keys(bucket, prefix="", suffix="", session=None):
    """
    Generate the keys in an S3 bucket.

    :param bucket: Name of the S3 bucket.
    :param prefix: Only fetch keys that start with this prefix (optional).
    :param suffix: Only fetch keys that end with this suffix (optional).
    """
    for obj in get_matching_s3_objects(bucket, prefix, suffix, session):
        yield obj["Key"]



"""Execute the AWS CLI update-kubeconfig command."""
from __future__ import print_function
import os
import subprocess
import logging
import sys

from stacker.lookups.handlers.output import OutputLookup
from runway.util import which

LOGGER = logging.getLogger(__name__)


def aws_eks_update_kubeconfig(provider, context, **kwargs):  # noqa pylint: disable=unused-argument
    """Execute the aws cli eks update-kubeconfig command.

    Args:
        provider (:class:`stacker.providers.base.BaseProvider`): provider
            instance
        context (:class:`stacker.context.Context`): context instance

    Returns: boolean for whether or not the hook succeeded.

    """
    eks_cluster_name = OutputLookup.handle(
        "%s::EksClusterName" % kwargs['stack'],
        provider=provider,
        context=context
    )
    LOGGER.info('writing kubeconfig...')
    subprocess.check_output(['runway', 'run-aws', '--', 'eks',
                             'update-kubeconfig', '--name', eks_cluster_name])
    LOGGER.info('kubeconfig written successfully...')

    # The newly-generated kubeconfig will have introduced a dependency on the
    # awscli. This is fine if a recent version is installed, or it's invoked
    # in a virtualenv with runway
    if not os.environ.get('PIPENV_ACTIVE') and (
            not os.environ.get('VIRTUAL_ENV') and not which('aws')):
        print('', file=sys.stderr)
        print('Warning: the generated kubeconfig uses the aws-cli for '
              'authentication, but it is not found in your environment. '
              'Either install it, or update the kubeconfig to use runway '
              'instead, e.g.:', file=sys.stderr)
        print('', file=sys.stderr)
        print('```', file=sys.stderr)
        print('      args:', file=sys.stderr)
        print('        - --region', file=sys.stderr)
        print('        - REGIONHERE', file=sys.stderr)
        print('        - eks', file=sys.stderr)
        print('        - get-token', file=sys.stderr)
        print('        - --cluster-name', file=sys.stderr)
        print('        - CLUSTERNAME', file=sys.stderr)
        print('        command: aws', file=sys.stderr)
        print('```', file=sys.stderr)
        print('', file=sys.stderr)
        print('becomes:', file=sys.stderr)
        print('', file=sys.stderr)
        print('```', file=sys.stderr)
        print('      args:', file=sys.stderr)
        print('        - run-aws', file=sys.stderr)
        print('        - --', file=sys.stderr)
        print('        - --region', file=sys.stderr)
        print('        - REGIONHERE', file=sys.stderr)
        print('        - eks', file=sys.stderr)
        print('        - get-token', file=sys.stderr)
        print('        - --cluster-name', file=sys.stderr)
        print('        - CLUSTERNAMEHERE', file=sys.stderr)
        print('        command: runway', file=sys.stderr)
        print('```', file=sys.stderr)
        print('', file=sys.stderr)
    return True



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
import re

from troposphere import Ref, GetAtt


HELPERS = {
    "Ref": Ref,
    "Fn::GetAtt": GetAtt
}

split_string = "(" + "|".join([r"%s\([^)]+\)" % h for h in HELPERS]) + ")"
replace_string = \
    r"(?P<helper>%s)\((?P<args>['\"]?[^)]+['\"]?)+\)" % '|'.join(HELPERS)

split_re = re.compile(split_string)
replace_re = re.compile(replace_string)


def cf_tokenize(s):
    """ Parses UserData for Cloudformation helper functions.

    http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/user-data.html
    http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference.html
    http://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/quickref-cloudformation.html#scenario-userdata-base64

    It breaks apart the given string at each recognized function (see HELPERS)
    and instantiates the helper function objects in place of those.

    Returns a list of parts as a result. Useful when used with Join() and
    Base64() CloudFormation functions to produce user data.

    ie: Base64(Join('', cf_tokenize(userdata_string)))
    """
    t = []
    parts = split_re.split(s)
    for part in parts:
        cf_func = replace_re.search(part)
        if cf_func:
            args = [a.strip("'\" ") for a in cf_func.group("args").split(",")]
            t.append(HELPERS[cf_func.group("helper")](*args).data)
        else:
            t.append(part)
    return t



"""K8s (kustomize) module."""

import logging
import os
import subprocess
import sys

import six

from runway.module import RunwayModule, run_module_command
from runway.env_mgr.kbenv import KBEnvManager
from runway.util import which

LOGGER = logging.getLogger('runway')


def gen_overlay_dirs(environment, region):
    """Generate possible overlay directories."""
    return [
        # Give preference to explicit environment-region dirs
        "%s-%s" % (environment, region),
        # Fallback to environment name only
        environment
    ]


def get_module_defined_k8s_ver(k8s_version_opts, env_name):
    """Return version of Terraform requested in module options."""
    if isinstance(k8s_version_opts, six.string_types):
        return k8s_version_opts
    if k8s_version_opts.get(env_name):
        return k8s_version_opts.get(env_name)
    if k8s_version_opts.get('*'):
        return k8s_version_opts.get('*')
    return None


def get_overlay_dir(overlays_path, environment, region):
    """Determine overlay directory to use."""
    for name in gen_overlay_dirs(environment, region):
        overlay_dir = os.path.join(overlays_path, name)
        if os.path.isfile(os.path.join(overlay_dir, 'kustomization.yaml')):
            return overlay_dir
    return overlay_dir  # fallback to last dir


def generate_response(overlay_path, module_path, environment, region):
    """Determine if environment is defined."""
    configfile = os.path.join(overlay_path, 'kustomization.yaml')
    if os.path.isdir(overlay_path) and os.path.isfile(configfile):
        LOGGER.info("Processing kustomize overlay: %s", configfile)
        return {'skipped_configs': False}
    LOGGER.error("No kustomize overlay for this environment/region found -- "
                 "looking for one of \"%s\"",
                 ', '.join(
                     [os.path.join(module_path, 'overlays', i, 'kustomization.yaml')  # noqa
                      for i in gen_overlay_dirs(environment, region)]
                 ))
    return {'skipped_configs': True}


class K8s(RunwayModule):
    """Kubectl Runway Module."""

    def run_kubectl(self, command='plan'):
        """Run kubectl."""
        kustomize_config_path = os.path.join(
            self.path,
            'overlays',
            get_overlay_dir(os.path.join(self.path,
                                         'overlays'),
                            self.context.env_name,
                            self.context.env_region)
        )
        response = generate_response(kustomize_config_path,
                                     self.path,
                                     self.context.env_name,
                                     self.context.env_region)
        if response['skipped_configs']:
            return response

        module_defined_k8s_ver = get_module_defined_k8s_ver(
            self.options.get('options', {}).get('kubectl_version', {}),
            self.context.env_name
        )
        if module_defined_k8s_ver:
            k8s_bin = KBEnvManager(self.path).install(module_defined_k8s_ver)
        elif os.path.isfile(os.path.join(kustomize_config_path,
                                         '.kubectl-version')):
            k8s_bin = KBEnvManager(kustomize_config_path).install()
        elif os.path.isfile(os.path.join(self.path,
                                         '.kubectl-version')):
            k8s_bin = KBEnvManager(self.path).install()
        elif os.path.isfile(os.path.join(self.context.env_root,
                                         '.kubectl-version')):
            k8s_bin = KBEnvManager(self.context.env_root).install()
        else:
            if not which('kubectl'):
                LOGGER.error('kubectl not available (a '
                             '".kubectl-version" file is not present '
                             'and "kubectl" not found in path). Fix '
                             'this by writing a desired kubectl version '
                             'to your module\'s .kubectl-version file '
                             'or installing kubectl.')
                sys.exit(1)
            k8s_bin = 'kubectl'

        kustomize_cmd = [k8s_bin, 'kustomize', kustomize_config_path]
        kustomize_yml = subprocess.check_output(kustomize_cmd,
                                                env=self.context.env_vars)
        if isinstance(kustomize_yml, bytes):  # python3 returns encoded bytes
            kustomize_yml = kustomize_yml.decode()
        if command == 'plan':
            LOGGER.info('The following yaml was generated by '
                        'kubectl:\n\n%s', kustomize_yml)
        else:
            LOGGER.debug('The following yaml was generated by '
                         'kubectl:\n\n%s', kustomize_yml)
        if command in ['apply', 'delete']:
            kubectl_command = [k8s_bin, command]
            if command == 'delete':
                kubectl_command.append('--ignore-not-found=true')
            kubectl_command.extend(['-k', kustomize_config_path])

            LOGGER.info('Running kubectl %s ("%s")...',
                        command,
                        ' '.join(kubectl_command))
            run_module_command(kubectl_command, self.context.env_vars)
        return response

    def plan(self):
        """Run kustomize build and display generated plan."""
        self.run_kubectl(command='plan')

    def deploy(self):
        """Run kubectl apply."""
        self.run_kubectl(command='apply')

    def destroy(self):
        """Run kubectl delete."""
        self.run_kubectl(command='delete')
